from typing import (
    Any,
    Callable,
    ClassVar,
    Generic,
    Mapping,
    NamedTuple,
    Self,
    TypeVar,
    TypedDict,
)

from loguru import logger
from pydantic import BaseModel
from qdrant_client import QdrantClient, models as qmodels
from qdrant_client.conversions import common_types as types

from .dataclass import DataClass
from .index.vectors import (
    SparseVectorType,
    BaseVectorIndex,
    DenseVectorType,
    DenseMultiVectorType,
    SparseVectorType,
    SparseVector,
    Vector,
    MultiVector,
)
from .index.payload import (
    BasePayloadIndex,
    Keyword,
    Integer,
    Text,
    Bool,
    Uuid,
    Geo,
    MultiGeo,
    Float,
    Datetime,
    KeywordType,
    IntegerType,
    FloatType,
    TextType,
    BoolType,
    UuidType,
    GeoType,
    MultiGeoType,
    DatetimeType,
)

T = TypeVar("T", bound=int | str)


class CollectionConfig(DataClass):
    collection_name: str | None = None
    shard_number: int | None = None
    sharding_method: types.ShardingMethod | None = None
    replication_factor: int | None = None
    write_consistency_factor: int | None = None
    on_disk_payload: bool | None = None
    hnsw_config: types.HnswConfigDiff | None = None
    optimizers_config: types.OptimizersConfigDiff | None = None
    wal_config: types.WalConfigDiff | None = None
    quantization_config: types.QuantizationConfig | None = None
    init_from: types.InitFrom | None = None
    timeout: int | None = None


class VectorConfigs(TypedDict):
    vectors_config: Mapping[str, types.VectorParams]
    sparse_vectors_config: Mapping[str, types.SparseVectorParams]


class PayloadParams(NamedTuple):
    params: BaseModel
    key: str | None


class IndexConfig(NamedTuple):
    vectors: VectorConfigs
    payloads: Mapping[str, PayloadParams]


class PointModel(DataClass, Generic[T]):
    __client__: ClassVar[QdrantClient]
    __collection_name__: ClassVar[str]
    __non_payload_fields__: ClassVar[set[str]] = {"id"}

    collection_config: ClassVar[CollectionConfig] = CollectionConfig()

    id: T

    @classmethod
    def _build_index_config(cls) -> VectorConfigs:
        vectors_config, sparse_vectors_config = {}, {}
        payload_indicies = {}

        for field, type_ in cls.__annotations__.items():
            if not hasattr(cls, field):
                continue

            # value = getattr(cls, field)

            # if isinstance(value, BasePayloadIndex):
            #     pass
            #     # msg = f"Payload field {field} must be of type {{}}. Got {type_}"

            #     # if isinstance(value, Keyword):
            #     #     assert type_ == KeywordType, msg.format(KeywordType)
            #     #     payload_indicies[field] = value.params
            #     # elif isinstance(value, Integer):
            #     #     assert type_ == IntegerType, msg.format(IntegerType)
            #     #     payload_indicies[field] = value.params
            #     # elif isinstance(value, Text):
            #     #     assert type_ == TextType, msg.format(TextType)
            #     #     payload_indicies[field] = value.params

            #     # elif isinstance(value, Uuid):
            #     #     assert type_ == UuidType, msg.format(UuidType)
            #     #     payload_indicies[field] = value.params
            #     # elif isinstance(value, Geo):
            #     #     assert type_ == GeoType, msg.format(GeoType)
            #     #     payload_indicies[field] = value.params
            #     # elif isinstance(value, MultiGeo):
            #     #     assert type_ == MultiGeoType, msg.format(MultiGeoType)
            #     #     payload_indicies[field] = value.params
            #     # elif isinstance(value, Float):
            #     #     assert type_ == FloatType, msg.format(FloatType)
            #     #     payload_indicies[field] = value.params

            # elif isinstance(value, Callable) and value.__name__ == "Bool":
            #     msg = f"Payload field {field} must be of type {{}}. Got {type_}"
            #     assert type_ == BoolType, msg.format(BoolType)
            #     payload_indicies[field] = value

            # elif isinstance(cls, BaseVectorIndex):
            msg = f"Vector field {field} must be of type {{}}. Got {type_}"
            value = getattr(cls, field)

            if isinstance(value, Vector):
                assert type_ == DenseVectorType, msg.format(DenseVectorType)
                vectors_config[field] = value.params
            elif isinstance(value, MultiVector):
                assert type_ == DenseMultiVectorType, msg.format(DenseMultiVectorType)
                vectors_config[field] = value.params
            elif isinstance(value, SparseVector):
                assert type_ == SparseVectorType, msg.format(SparseVectorType)
                sparse_vectors_config[field] = value.params

        return {
            "vectors_config": vectors_config,
            "sparse_vectors_config": sparse_vectors_config,
        }

    @classmethod
    def init_collection(cls, client: QdrantClient):
        cls.__client__ = client

        if cls.collection_config.collection_name is None:
            cls.collection_config.collection_name = cls.__name__

        cls.__collection_name__ = cls.collection_config.collection_name

        index_config = cls._build_index_config()

        cls.__non_payload_fields__.update(index_config["vectors_config"])
        cls.__non_payload_fields__.update(index_config["sparse_vectors_config"])

        if client.collection_exists(cls.collection_config.collection_name):
            pass
            # logger.info(
            #     "Collection already exists, if you want to update it use update_collection manually",
            # )
        else:
            index_config |= cls.collection_config.to_dict()
            client.create_collection(**index_config)  # type: ignore

    @classmethod
    def _from_record(
        cls, record: types.Record | types.ScoredPoint, set_persisted: bool = False
    ) -> Self:
        point = cls(id=record.id, **record.payload or {}, **record.vector or {})  # type: ignore
        point._persisted = set_persisted
        return point

    def __post_init__(self):
        self._persisted = False
        self._current_prefetch: types.Prefetch | None = None
        self._context_prefetch_is_on = False

    @property
    def persisted(self) -> bool:
        return self._persisted

    def payload(self) -> dict[str, Any]:
        return self.to_dict(exclude=self.__non_payload_fields__)

    def vectors(self) -> dict[str, types.Vector]:
        result = {}

        for field in self.__non_payload_fields__:
            if field == "id":
                continue

            if vector := getattr(self, field, None):
                if type(vector) is SparseVectorType:
                    result[field] = qmodels.SparseVector(
                        indices=vector[0], values=vector[1]
                    )
                else:
                    result[field] = vector

        return result


def init_models(client: QdrantClient, models: list[type[PointModel[T]]]):
    for model in models:
        model.init_collection(client)



from typing import Mapping, TypeAlias, TypedDict

from qdrant_client import models
from qdrant_client.conversions import common_types as types


DenseVectorType: TypeAlias = list[float]
DenseMultiVectorType: TypeAlias = list[DenseVectorType]
SparseVectorType: TypeAlias = tuple[list[int], list[float]]
BaseVectorType = DenseVectorType | DenseMultiVectorType | SparseVectorType


class BaseVectorIndex:
    _params: types.VectorParams | types.SparseVectorParams

    @property
    def params(self):
        return self._params


class Vector(DenseVectorType, BaseVectorIndex):
    """
    Params of single vector data storage
    """

    def __init__(
        self,
        size: int,
        distance: models.Distance,
        hnsw_config: models.HnswConfigDiff | None = None,
        quantization_config: models.QuantizationConfig | None = None,
        on_disk: bool | None = None,
        datatype: models.Datatype | None = None,
    ):
        self._params = models.VectorParams(
            size=size,
            distance=distance,
            hnsw_config=hnsw_config,
            quantization_config=quantization_config,
            on_disk=on_disk,
            datatype=datatype,
        )


class MultiVector(DenseMultiVectorType, BaseVectorIndex):
    """
    Params of multi vector data storage
    """

    def __init__(
        self,
        single_size: int,
        distance: models.Distance,
        multivector_config: models.MultiVectorConfig = models.MultiVectorConfig(
            comparator=models.MultiVectorComparator.MAX_SIM
        ),
        hnsw_config: models.HnswConfigDiff | None = None,
        quantization_config: models.QuantizationConfig | None = None,
        on_disk: bool | None = None,
        datatype: models.Datatype | None = None,
    ):
        self._params = models.VectorParams(
            size=single_size,
            distance=distance,
            hnsw_config=hnsw_config,
            quantization_config=quantization_config,
            on_disk=on_disk,
            datatype=datatype,
            multivector_config=multivector_config,
        )


class SparseVector(SparseVectorType, BaseVectorIndex):
    """
    Params of single sparse vector data storage
    """

    def __init__(
        self,
        index: models.SparseIndexParams | None = None,
        modifier: models.Modifier | None = None,
    ):
        self._params = models.SparseVectorParams(index=index, modifier=modifier)


# class VectorConfigs(TypedDict):
#     vectors_config: Mapping[str, types.VectorParams]
#     sparse_vectors_config: Mapping[str, types.SparseVectorParams]


# def build_vectors_config(cls: type[object]) -> VectorConfigs:
#     vectors_config, sparse_vectors_config = {}, {}

#     for field, type_ in cls.__annotations__.items():
#         if not hasattr(cls, field):
#             continue

#         msg = f"Vector field {field} must be of type {{}}. Got {type_}"
#         vector = getattr(cls, field)
#         if isinstance(vector, Vector):
#             assert type_ == DenseVectorType, msg.format(DenseVectorType)
#             vectors_config[field] = vector.params
#         elif isinstance(vector, MultiVector):
#             assert type_ == DenseMultiVectorType, msg.format(DenseMultiVectorType)
#             vectors_config[field] = vector.params
#         elif isinstance(vector, SparseVector):
#             assert type_ == SparseVectorType, msg.format(SparseVectorType)
#             sparse_vectors_config[field] = vector.params

#     return {
#         "vectors_config": vectors_config,
#         "sparse_vectors_config": sparse_vectors_config,
#     }



from .crud import CRUDPoint as PointModel, ReadOptions, WriteOptions
from .model import CollectionConfig, init_models

__all__ = [
    "PointModel",
    "ReadOptions",
    "WriteOptions",
    "CollectionConfig",
    "init_models",
]



from typing import Any, ClassVar, dataclass_transform


class Serializable:
    def __init__(self, **kwargs: Any):
        for field in self.fields:
            setattr(self, field, kwargs.get(field))

    @property
    def fields(self) -> dict[str, Any]:
        parent_annotations = {}
        parents_stack = [self.__class__.__bases__]

        while parents_stack:
            for base in parents_stack.pop():
                if hasattr(base, "__annotations__"):
                    for field, type_ in base.__annotations__.items():
                        if getattr(type_, "__origin__", None) is not ClassVar:
                            parent_annotations[field] = type_
                if base.__bases__:
                    parents_stack.append(base.__bases__)

        return parent_annotations | type(self).__annotations__

    def to_dict(self, exclude: set[str] | None = None) -> dict[str, Any]:
        return {
            field: getattr(self, field)
            for field in self.fields
            if exclude is None or field not in exclude
        }


@dataclass_transform(kw_only_default=True)
class DataClass(Serializable):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.__post_init__()

    def __post_init__(self):
        pass



from typing import Any, TypeAlias

from qdrant_client import models

KeywordType: TypeAlias = list[str]
IntegerType: TypeAlias = int
FloatType: TypeAlias = float
TextType: TypeAlias = str
BoolType: TypeAlias = bool
UuidType: TypeAlias = str
GeoType: TypeAlias = tuple[float, float]
MultiGeoType: TypeAlias = list[tuple[float, float]]
DatetimeType: TypeAlias = str


class BasePayloadIndex:
    _params: Any = None
    _key = None

    @property
    def params(self) -> Any:
        return self._params

    @property
    def key(self) -> str | None:
        return self._key


class Keyword(list[str], BasePayloadIndex):
    def __init__(
        self,
        is_tenant: bool | None = None,
        on_disk: bool | None = None,
        key: str | None = None,
    ):
        """
        Keyword index

        :param is_tenant: If true - used for tenant optimization. Default: false.
        :param on_disk: If true, store the index on disk. Default: false.
        :param key: Key for nested field. You can use dot notation to specify the path to the nested field.
        """
        self._key = key
        self._params = models.KeywordIndexParams(
            type=models.KeywordIndexType.KEYWORD,
            is_tenant=is_tenant,
            on_disk=on_disk,
        )


class Integer(int, BasePayloadIndex):
    def __init__(
        self,
        lookup: bool | None = None,
        range: bool | None = None,
        is_principal: bool | None = None,
        on_disk: bool | None = None,
        key: str | None = None,
    ):
        """
        Integer index

        :param lookup: If true - support direct lookups.
        :param range: If true - support ranges filters.
        :param is_principal: If true - use this key to organize storage of the collection data. This option assumes that this key will be used in majority of filtered requests.
        :param on_disk: If true, store the index on disk. Default: false.
        :param key: Key for nested field. You can use dot notation to specify the path to the nested field.
        """
        self._key = key
        self._index_params = models.IntegerIndexParams(
            type=models.IntegerIndexType.INTEGER,
            lookup=lookup,
            range=range,
            is_principal=is_principal,
            on_disk=on_disk,
        )


class Float(float, BasePayloadIndex):
    def __init__(
        self,
        is_principal: bool | None = None,
        on_disk: bool | None = None,
        key: str | None = None,
    ):
        """
        Float index

        :param is_principal: If true - use this key to organize storage of the collection data. This option assumes that this key will be used in majority of filtered requests.
        :param on_disk: If true, store the index on disk. Default: false.
        :param key: Key for nested field. You can use dot notation to specify the path to the nested field.
        """
        self._key = key
        self._index_params = models.FloatIndexParams(
            type=models.FloatIndexType.FLOAT,
            is_principal=is_principal,
            on_disk=on_disk,
        )


def Bool(key: str | None = None) -> Any:  # type: ignore
    return key, models.BoolIndexParams(
        type=models.BoolIndexType.BOOL,
    )


class Geo(tuple[float, float], BasePayloadIndex):
    def __init__(self, on_disk: bool | None = None, key: str | None = None):
        """
        Geo index

        :param on_disk: If true, store the index on disk. Default: false.
        :param key: Key for nested field. You can use dot notation to specify the path to the nested field.
        """
        self._key = key
        self._index_params = models.GeoIndexParams(
            type=models.GeoIndexType.GEO,
            on_disk=on_disk,
        )


class MultiGeo(list[tuple[float, float]], BasePayloadIndex):
    def __init__(self, on_disk: bool | None = None, key: str | None = None):
        """
        Multiple geo index

        :param on_disk: If true, store the index on disk. Default: false.
        :param key: Key for nested field. You can use dot notation to specify the path to the nested field.
        """
        self._key = key
        self._index_params = models.GeoIndexParams(
            type=models.GeoIndexType.GEO,
            on_disk=on_disk,
        )


class Datetime(int, BasePayloadIndex):
    def __init__(
        self,
        is_principal: bool | None = None,
        on_disk: bool | None = None,
        key: str | None = None,
    ):
        """
        Datetime index

        :param is_principal: If true - use this key to organize storage of the collection data. This option assumes that this key will be used in majority of filtered requests.
        :param on_disk: If true, store the index on disk. Default: false.
        :param key: Key for nested field. You can use dot notation to specify the path to the nested field.
        """
        self._key = key
        self._index_params = models.DatetimeIndexParams(
            type=models.DatetimeIndexType.DATETIME,
            is_principal=is_principal,
            on_disk=on_disk,
        )


class Text(str, BasePayloadIndex):
    def __init__(
        self,
        tokenizer: models.TokenizerType = models.TokenizerType.WORD,
        min_token_len: int | None = None,
        max_token_len: int | None = None,
        lowercase: bool | None = None,
        on_disk: bool | None = None,
        key: str | None = None,
    ):
        """
        Full text index

        :param tokenizer: Tokenizer type.
        :param min_token_len: Minimum characters to be tokenized.
        :param max_token_len: Maximum characters to be tokenized.
        :param lowercase: If true, lowercase all tokens. Default: true.
        :param on_disk: If true, store the index on disk. Default: false.
        :param key: Key for nested field. You can use dot notation to specify the path to the nested field.
        """
        self._key = key
        self._index_params = models.TextIndexParams(
            type=models.TextIndexType.TEXT,
            tokenizer=tokenizer,
            min_token_len=min_token_len,
            max_token_len=max_token_len,
            lowercase=lowercase,
            on_disk=on_disk,
        )


class Uuid(str, BasePayloadIndex):
    def __init__(
        self,
        is_tenant: bool | None = None,
        on_disk: bool | None = None,
        key: str | None = None,
    ):
        """
        Uuid index

        :param is_tenant: If true - used for tenant optimization. Default: false.
        :param on_disk: If true, store the index on disk. Default: false.
        :param key: Key for nested field. You can use dot notation to specify the path to the nested field.
        """
        self._key = key
        self._index_params = models.UuidIndexParams(
            type=models.UuidIndexType.UUID,
            is_tenant=is_tenant,
            on_disk=on_disk,
        )



from .payload import Keyword, Integer, Text, Bool, Uuid, Geo, MultiGeo, Float, Datetime
from .vectors import Vector, MultiVector, SparseVector

__all__ = [
    "Keyword",
    "Integer",
    "Text",
    "Bool",
    "Uuid",
    "Geo",
    "MultiGeo",
    "Float",
    "Datetime",
    "Vector",
    "MultiVector",
    "SparseVector",
]



from typing import Iterable, Iterator, NamedTuple, Self, Sequence
from types import TracebackType

from qdrant_client import models
from qdrant_client.conversions import common_types as types

from .model import PointModel, T


class ReadOptions(NamedTuple):
    with_vectors: bool | Sequence[str] = False
    consistency: types.ReadConsistency | None = None
    shard_key_selector: types.ShardKeySelector | None = None
    timeout: int | None = None


class WriteOptions(NamedTuple):
    wait: bool = True
    ordering: types.WriteOrdering | None = None
    shard_key_selector: types.ShardKeySelector | None = None


class CRUDPoint(PointModel[T]):
    @classmethod
    def get(
        cls,
        id: T,
        read_options: ReadOptions = ReadOptions(),
    ) -> Self:
        """
        Get a point from Qdrant.

        Args:
            id (T): The id of the point to get.
            read_options (ReadOptions, optional): Read options. Defaults to ReadOptions().

        Returns:
            Self: The point.
        """
        record, *_ = cls.__client__.retrieve(
            cls.__collection_name__, ids=[id], **read_options._asdict()
        )
        return cls._from_record(record, set_persisted=True)

    @classmethod
    def scroll(
        cls,
        scroll_filter: types.Filter | None = None,
        limit: int = 10,
        order_by: types.OrderBy | None = None,
        read_options: ReadOptions = ReadOptions(),
    ) -> Iterator[list[Self]]:
        """
        Scroll points from Qdrant.

        Args:
            scroll_filter (types.Filter | None, optional): Filter to apply. Defaults to None.
            limit (int, optional): Number of points to fetch per scroll. Defaults to 10.
            order_by (types.OrderBy | None, optional): Order by. Defaults to None.
            read_options (ReadOptions, optional): Read options. Defaults to ReadOptions().

        Yields:
            Iterator[list[Self]]: Iterator of lists of points.
        """
        offset = None

        while True:
            records, offset = cls.__client__.scroll(
                cls.__collection_name__,
                scroll_filter=scroll_filter,
                offset=offset,
                limit=limit,
                order_by=order_by,
                **read_options._asdict(),
            )
            yield [cls._from_record(record, set_persisted=True) for record in records]

            if offset is None:
                break

    @classmethod
    def insert_many(
        cls,
        *points: Self,
        write_options: WriteOptions = WriteOptions(),
    ) -> None:
        """
        Save the points to Qdrant.

        Args:
            write_options (WriteOptions, optional): Write options. Defaults to WriteOptions().
        """
        cls.__client__.upsert(
            cls.__collection_name__,
            points=[
                models.PointStruct(
                    id=point.id, payload=point.payload(), vector=point.vectors()
                )
                for point in points
            ],
            **write_options._asdict(),
        )

    @classmethod
    def count(
        cls,
        count_filter: types.Filter | None = None,
        exact: bool = True,
        shard_key_selector: types.ShardKeySelector | None = None,
        timeout: int | None = None,
    ) -> int:
        """
        Count points in collection.

        Args:
            count_filter (types.Filter | None, optional): Filter to apply. Defaults to None.
            exact (bool, optional): Whether to use exact count. Defaults to True.
            shard_key_selector (types.ShardKeySelector | None, optional): Shard key selector. Defaults to None.
            timeout (int | None, optional): Timeout. Defaults to None.

        Returns:
            int: Number of points
        """
        result = cls.__client__.count(
            cls.__collection_name__,
            count_filter=count_filter,
            exact=exact,
            shard_key_selector=shard_key_selector,
            timeout=timeout,
        )
        return result.count

    def save(
        self,
        overwrite_vectors: bool = False,
        write_options: WriteOptions = WriteOptions(),
    ) -> None:
        """
        Save the point to Qdrant.

        Args:
            overwrite_vectors (bool, optional): Whether to overwrite the vectors if they already exist. Defaults to False.
            write_options (WriteOptions, optional): Write options. Defaults to WriteOptions().
        """
        write_kwargs = write_options._asdict()
        client = self.__client__
        collection_name = self.__collection_name__
        payload = self.payload()

        if self._persisted:
            client.set_payload(
                collection_name,
                payload=payload,
                points=[self.id],
                **write_kwargs,
            )

            if overwrite_vectors:
                client.update_vectors(
                    collection_name,
                    points=[models.PointVectors(id=self.id, vector=self.vectors())],
                    **write_kwargs,
                )
        else:
            client.upsert(
                collection_name,
                points=[
                    models.PointStruct(
                        id=self.id, payload=payload, vector=self.vectors()
                    )
                ],
                **write_kwargs,
            )
            self._persisted = True

    def delete(self, write_options: WriteOptions = WriteOptions()) -> None:
        """
        Delete the point from Qdrant.

        Args:
            write_options (WriteOptions, optional): Write options. Defaults to WriteOptions().
        """
        if not self._persisted:
            raise ValueError(
                "Cannot delete non-persisted point. You need to save it first."
            )

        self.__client__.delete(
            self.__collection_name__,
            points_selector=[self.id],
            **write_options._asdict(),
        )

    def sync(self, read_options: ReadOptions = ReadOptions()) -> None:
        """
        Syncronize the object with Qdrant record.

        Args:
            read_options (ReadOptions, optional): Read options. Defaults to ReadOptions().
        """
        persisted_point = self.get(self.id, read_options)
        for field in self.fields:
            if persisted_value := getattr(persisted_point, field, None):
                setattr(self, field, persisted_value)
        self._persisted = True

    def prefetch(
        self,
        using: str,
        limit: int | None = None,
        score_threshold: float | None = None,
        prefetch_filter: types.Filter | None = None,
        params: types.SearchParams | None = None,
    ) -> Self:
        """
        Prefetch points from Qdrant.

        Args:
            using (str): The using vector to use.
            limit (int | None, optional): The limit of points to prefetch. Defaults to None.
            score_threshold (float | None, optional): The score threshold to use. Defaults to None.
            prefetch_filter (types.Filter | None, optional): The filter to use. Defaults to None.
            params (types.SearchParams | None, optional): The search params to use. Defaults to None.
        """

        # if not self._context_prefetch_is_on:
        #     raise ValueError(
        #         "Prefetch must be used in context manager. "
        #         "Use `with point.prefetch(...):`"
        #     )

        self._context_prefetch_is_on = False

        prefetch_query = models.Prefetch(
            query=self.id,
            using=using,
            limit=limit,
            score_threshold=score_threshold,
            filter=prefetch_filter,  # type: ignore
            prefetch=self._current_prefetch,
            params=params,  # type: ignore
        )

        self._current_prefetch = prefetch_query
        return self

    def __enter__(self) -> Self:
        return self

    def __exit__(self, exc_type: type, exc_value: Exception, traceback: TracebackType):
        if current_prefetch := self._current_prefetch:
            self._current_prefetch = current_prefetch.prefetch  # type: ignore

    def neighbors(
        self,
        using: str,
        limit: int = 10,
        score_threshold: float | None = None,
        query_filter: types.Filter | None = None,
        read_options: ReadOptions = ReadOptions(),
        search_params: types.SearchParams | None = None,
    ) -> list[tuple[Self, float]]:
        """
        Get neighbors for the point.

        Args:
            using (str): which vector field to use
            limit (int, optional): Limit. Defaults to 10.
            score_threshold (float | None, optional): Score threshold. Defaults to None.
            query_filter (types.Filter | None, optional): Query filter. Defaults to None.
            prefetch (models.Prefetch | None, optional): Prefetch. Defaults to None.
            read_options (ReadOptions, optional): Read options. Defaults to ReadOptions().
            search_params (SearchParams, optional): Search params.

        Yields:
            Iterable[tuple[Self, float]]: Iterator of neighbors.
        """

        if not self._persisted:
            raise ValueError(
                "Cannot get neighbors for non-persisted point. You need to save it first."
            )

        response = self.__client__.query_points(
            self.__collection_name__,
            query=self.id,
            using=using,
            limit=limit,
            score_threshold=score_threshold,
            query_filter=query_filter,
            prefetch=self._current_prefetch,
            search_params=search_params,
            **read_options._asdict(),
        )

        return [
            (self._from_record(record, set_persisted=True), record.score)
            for record in response.points
        ]
